#
# coding=utf-8
#*********************************************************************
#
# MPDCLient-Twisted
#
# twisted-port of mpdclient3 (C) 2005 Nick Welch
#
# written by Timo Boettcher
# pympdtouchgui (near) spida (.) net
#
# (C) 2009 Timo Boettcher
# This is free software. You may redistribute copies of it under the
# terms of the GNU General Public License (Version 2)
# <http://www.gnu.org/licenses/gpl.html>.
# There is NO WARRANTY, to the extent permitted by law.
#
#*********************************************************************

try:
	import sys
	from twisted.internet.protocol import ClientFactory
	from twisted.internet import protocol, defer
	from twisted.protocols.basic import LineOnlyReceiver
	from twisted.protocols import policies
except ImportError, err:
	print "couldn't load module. %s"%(err)
	sys.exit(2)

ZERO = 0
ONE = 1
MANY = 2

plitem_delim = ["file", "directory", "playlist"]
litem_delim = ["album", "artist", "title", "track", "name", "genre", "date", "composer", "performer", "comment" ] #For use in list command.

class MPDClient(LineOnlyReceiver, policies.TimeoutMixin):
	_commands = {
		# (name, nargs): (format string, nresults, results_type_name, delimiter_key, available)

		# delimiter key is for commands that return multiple results.  we use this
		# string to detect the beginning of a new result object.

		# if results_type_name is empty, the result object's .type will be set to
		# the key of the first key/val pair in it; otherwise, it will be set to
		# results_type_name.

		("kill",          0): ('%s', ZERO, '', [], False),
		("outputs",       0): ('%s', MANY, 'outputs', ['outputid'], False),
		("clear",         0): ('%s', ZERO, '', [], False),
		("currentsong",   0): ('%s', ONE, '', [], False),
		("shuffle",       0): ('%s', ZERO, '', [], False),
		("next",          0): ('%s', ZERO, '', [], False),
		("previous",      0): ('%s', ZERO, '', [], False),
		("stop",          0): ('%s', ZERO, '', [], False),
		("clearerror",    0): ('%s', ZERO, '', [], False),
		("close",         0): ('%s', ZERO, '', [], False),
		("commands",      0): ('%s', MANY, 'commands', ['command'], True),
		("notcommands",   0): ('%s', MANY, 'notcommands', ['command'], True),
		("ping",          0): ('%s', ZERO, '', [], False),
		("stats",         0): ('%s', ONE, 'stats', [], False),
		("status",        0): ('%s', ONE, 'status', [], False),
		("play",          0): ('%s', ZERO, '', [], False),
		("playlistinfo",  0): ('%s', MANY, '', plitem_delim, False),
		("playlistid",    0): ('%s', MANY, '', plitem_delim, False),
#		("getqueueinfo",  0): ('%s', MANY, '', plitem_delim, False),
		("lsinfo",        0): ('%s', MANY, '', plitem_delim, False),
		("update",        0): ('%s', ZERO, '', [], False),
		("listall",       0): ('%s', MANY, '', plitem_delim, False),
		("listallinfo",   0): ('%s', MANY, '', plitem_delim, False),

		("command_list_begin", 0): ('%s', ZERO, '', [], True),
		("command_list_end",   0): ('%s', ZERO, '', [], True),

		("disableoutput", 1): ("%s %d", ZERO, '', [], False),
		("enableoutput",  1): ("%s %d", ZERO, '', [], False),
		("delete",        1): ('%s %d', ZERO, '', [], False), # <int song>
		("deleteid",      1): ('%s %d', ZERO, '', [], False), # <int songid>
		("playlistinfo",  1): ('%s %d', MANY, '', plitem_delim, False), # <int song>
		("playlistid",    1): ('%s %d', MANY, '', plitem_delim, False), # <int songid>
#		("queue",         1): ('%s %d', ZERO, '', [], False), # <int songid>
#		("dequeue",       1): ('%s %d', ZERO, '', [], False), # <int songid>
		("crossfade",     1): ('%s %d', ZERO, '', [], False), # <int seconds>
		("play",          1): ('%s %d', ZERO, '', [], False), # <int song>
		("playid",        1): ('%s %d', ZERO, '', [], False), # <int songid>
		("random",        1): ('%s %d', ZERO, '', [], False), # <int state>
		("repeat",        1): ('%s %d', ZERO, '', [], False), # <int state>
		("setvol",        1): ('%s %d', ZERO, '', [], False), # <int vol>
		("plchanges",     1): ('%s %d', MANY, '', plitem_delim, False), # <playlist version>
		("pause",         1): ('%s %d', ZERO, '', [], False), # <bool pause>

		("update",      1): ('%s "%s"', ONE, 'update', [], False), # <string path>
		("listall",     1): ('%s "%s"', MANY, '', plitem_delim, False), # <string path>
		("listallinfo", 1): ('%s "%s"', MANY, '', plitem_delim, False), # <string path>
		("lsinfo",      1): ('%s "%s"', MANY, '', plitem_delim, False), # <string directory>
		("add",         1): ('%s "%s"', ZERO, '', [], False), # <string>
		("load",        1): ('%s "%s"', ZERO, '', [], False), # <string name>
		("rm",          1): ('%s "%s"', ZERO, '', [], False), # <string name>
		("save",        1): ('%s "%s"', ZERO, '', [], False), # <string playlist name>
		("password",    1): ('%s "%s"', ZERO, '', [], True), # <string password>

		("move",   2): ("%s %d %d", ZERO, '', [], False), # <int from> <int to>
		("moveid", 2): ("%s %d %d", ZERO, '', [], False), # <int songid from> <int to>
		("swap",   2): ("%s %d %d", ZERO, '', [], False), # <int song1> <int song2>
		("swapid", 2): ("%s %d %d", ZERO, '', [], False), # <int songid1> <int songid2>
		("seek",   2): ("%s %d %d", ZERO, '', [], False), # <int song> <int time>
		("seekid", 2): ("%s %d %d", ZERO, '', [], False), # <int songid> <int time>

		# <string type> <string what>
		("find",   2): ('%s "%s" "%s"', MANY, '', plitem_delim, False),

		# <string type> <string what>
		("search", 2): ('%s "%s" "%s"', MANY, '', plitem_delim, False),
		("search", 4): ('%s "%s" "%s" "%s" "%s"', MANY, '', plitem_delim, False),

		# list <metadata arg1> [<metadata arg2> <search term>]

		# <metadata arg1>
		("list", 1): ('%s "%s"', MANY, '', litem_delim, False),

		# <metadata arg1> <metadata arg2> <search term>
		("list", 3): ('%s "%s" "%s" "%s"', MANY, '', litem_delim, False),

		("idle",          0): ('%s', ONE, '', [], False),
		("idle",          1): ('%s %s', ONE, '', [], False),
		("idle",          2): ('%s %s %s', ONE, '', [], False),
		("idle",          3): ('%s %s %s %s', ONE, '', [], False),
		("idle",          4): ('%s %s %s %s %s', ONE, '', [], False),
		("idle",          5): ('%s %s %s %s %s %s', ONE, '', [], False),
		("idle",          6): ('%s %s %s %s %s %s %s', ONE, '', [], False),
		("idle",          7): ('%s %s %s %s %s %s %s %s', ONE, '', [], False),
	}
	delimiter = "\n"
	_states = ["disconnected", "connected", "unauthenticated", "password sent", "authenticated"]
	_existingevents = ["database", "stored_playlist", "playlist", "player", "mixer", "output", "options"]

	def __init__(self, password=None, ConnectCallback=None, ReadyCallback=None, DisconnectCallback=None):
		self._timeout = None
		self._state = "disconnected"
		self._idlestore = None #triple of <arguments, deferred, responsehandled>
		self._idleonce = True
		self._availablecommands = ""
		self._answer = [] # accumulate answer here
		self._commandqueue = [] #tuple of <command, args, deferred>
		self._currentcommand = None #tuple of <command, args, deferred>
		self._Debug = False

		self.password = password
		self.mpd_version = None

		# network callbacks
		self.CallbackConnect = ConnectCallback
		self.CallbackReady = ReadyCallback
		self.CallbackDisconnect = DisconnectCallback

		# idle callbacks (events from mpd-server)
		self.CallbackChanged = {}

	def setEventCallback(self, Event, Callback):
		if Event in self._existingevents:
			self.CallbackChanged[Event] = Callback

	def connectionMade(self):
		self._state = "connected"
		self.setTimeout(self._timeout)

	def connectionLost(self, reason=protocol.connectionDone):
		self._state = "disconnected"
		if self.CallbackDisconnect:
			self.CallbackDisconnect()

	def lineReceived(self, line):
		self.resetTimeout()

		if line.startswith("ACK") or line.find("you don't have permission") >= 0:
			self._handleError(line)
			return

		# simple state machine
		if self._state == "disconnected":
			raise RuntimeError("shouldn't receive anything in disconnected state")
		elif self._state == "connected":
			if line.startswith("OK MPD "):
				self._state = "unauthenticated"
				self.mpd_version = line[len("OK MPD "):].strip()
			else:
				raise RuntimeError("No MPD listening on the other side")
			if self.CallbackConnect:
				self.CallbackConnect()
			if self.password != "":
				self.Login()
			else:
				self.getAvailableCommands()
				if self.CallbackReady:
					self.CallbackReady()
		elif self._state == "unauthenticated":
			self._handleAnswer(line)
		elif self._state == "password sent":
			self._handlePasswordAnswer(line)
		elif self._state == "authenticated":
			self._handleAnswer(line)


	def _handleError(self, line):
		(command, arguments, deferred) = self._currentcommand
		self._currentcommand = None
		if self._state == "password sent":
			if line.endswith("incorrect password"):
				self._state = "unauthenticated"
				deferred.errback(Exception("Wrong Password"))
			else:
				deferred.errback(Exception("error "+ line.split(None, 1)[1]))
		else:
			if line.find("you don't have permission") >= 0:
				deferred.errback(Exception("Insufficient Permissions"))
			else:
				deferred.errback(Exception("error "+ line.split(None, 1)[1]))
		self._sendCommand() # next

	def _handleAnswer(self, line):
		if self._currentcommand:
			if line == "OK":
				# received complete result in _answer, now parse it
				if self._idlestore != None and not self._idlestore[2]:
					(arguments, deferred, responsehandled) = self._idlestore
					self._idlestore = (arguments, deferred, True)
				else:
					(command, arguments, deferred) = self._currentcommand
					self._currentcommand = None
					format, howmany, type, keywords, available = self._getCommand(command, arguments)
					if howmany == ZERO:
						result = None
					elif howmany == ONE:
						result = self._getAnswerOneField(self._answer, keywords, type)
					elif howmany == MANY:
						result = self._getAnswerManyFields(self._answer, keywords, type)
					else:
						RuntimeError("unknown responsecount")

					self._answer = []
					deferred.callback(result)
					self._sendCommand() # next
			else:
				# store incomplete answer for later parsing
				self._answer.append(line)
		else:
			# all commands were executed and answered so why did we receive this? we didn't ask!
			print "received %s: %s"%(self._state, line)

	def _handlePasswordAnswer(self, line):
		(command, arguments, deferred) = self._currentcommand
		if line == "OK":
			self._state = "authenticated"
			if command == "password":
				deferred.callback(defer.SUCCESS)
				self._currentcommand = None
				if self.CallbackReady:
					self.CallbackReady()
			else:
				raise RuntimeError("State machine fail")
		else:
			self._state = "unauthenticated"
			self._currentcommand = None
			deferred.errback(Exception("Unexpected response"))

	def _getAnswerOneField(self, answer, keywords, type):
		entity = dictobj()
		if type:
			entity['type'] = type
		for line in answer:
			pair = self._getPair(line)
			key, val = pair
			key = key.lower()
			if key in keywords and key in entity.keys():
				return entity
			if key in keywords:
				for item in keywords:
					if item in entity.keys():
						return entity

			if not type and 'type' not in entity.keys():
				entity['type'] = key

			entity[key] = val
		return entity

	def _getAnswerManyFields(self, answer, keywords, type):
		list = []
		entity = dictobj()
		if type:
			entity['type'] = type
		for line in answer:
			pair = self._getPair(line)
			key, val = pair
			key = key.lower()
			if key in keywords and key in entity.keys():
				list.append(entity)
				entity = dictobj()
			if key in keywords:
				for item in keywords:
					if item in entity.keys():
						list.append(entity)
						entity = dictobj()

			if not type and 'type' not in entity.keys():
				entity['type'] = key

			entity[key] = val
		list.append(entity)
		return list

	def _escape(self, text):
		# join/split is faster than replace
		text = '\\\\'.join(text.split('\\')) # \ -> \\
		text = '\\"'.join(text.split('"')) # " -> \"
		return text

	def _getCommand(self, command, arguments):
		try:
			return self._commands[(command, len(arguments))]
		except KeyError:
			raise RuntimeError("no such command: %s (%d args)"%(command, len(arguments)))

#	Not needed
#	def _isCommand(self, command):
#		return command in [ k[0] for k in self._commands.keys() ]
#
#	def _getCommandArgumentCount(self, command):
#		for k in self._commands.keys():
#			if k[0] == command:
#				return k[1]

	def _getPair(self, line):
		pair = line.split(": ", 1)
		if len(pair) != 2:
			raise RuntimeError("bogus response: ``%s''"%line)
		return pair

	def sendCommand(self, command, arguments=[]):
		deferred = defer.Deferred()
		commandtuple = (command, arguments, deferred)
		self._commandqueue.append(commandtuple)
		self._sendCommand()
		return deferred

	def _sendCommand(self):
		if self._currentcommand:
			# cancel idle first, if necessary
			(command, arguments, deferred) = self._currentcommand
			if (command == "idle") and self._commandqueue:
				self._idlestore = (arguments, deferred, False)
				if self._Debug:
					(command, arguments, deferred) = self._commandqueue[0]
				self.sendLine("noidle")
				self._getCommandFromQueue()
		else:
			if self._commandqueue:
				self._getCommandFromQueue()
			elif self._idlestore != None:
				# resume idle
				(arguments, deferred, responsehandled) = self._idlestore
				commandtuple = ("idle", arguments, deferred)
				self._commandqueue.append(commandtuple)
				self._getCommandFromQueue()
				# FIXME use _EventsToWait instead of stored arguments

	def _getCommandFromQueue(self):
		if self._commandqueue:
			if len(self._commandqueue) > 1:
				# find first non-idle command
				for number, commandtuple in enumerate(self._commandqueue):
					(command, arguments, deferred) = commandtuple
					if command != "idle":
						break
				commandtuple = self._commandqueue.pop(number)
				self._sendCommandReal(commandtuple)
			else:
				commandtuple = self._commandqueue.pop(0)
				self._sendCommandReal(commandtuple)

	def _sendCommandReal(self, commandtuple):
		self._currentcommand = commandtuple
		(command, arguments, deferred) = commandtuple
		if command == "password":
			self._state = "password sent"
		format = self._getCommand(command, arguments)[0]
		arguments = list(arguments[:])
		for i, argument in enumerate(arguments):
			if not isinstance(argument, int):
				arguments[i] = self._escape(str(argument))

		available = self._getCommand(command, arguments)[4]
		if available:
			self.sendLine(format%tuple([command] + list(arguments)))
		else:
			self._currentcommand = None
			deferred.errback(Exception("Command %s unavailable"%command))


	#################################################################


	def Login(self, password=None):
		if not password:
			password = self.password
		if password:
			deferred = self.sendCommand("password", [password])
			deferred.addCallback(self._loginCallback)

	def _loginCallback(self, result):
		self.getAvailableCommands()

	def getAvailableCommands(self):
		deferreda = self.sendCommand("commands")
		deferreda.addCallback(self._availableCommandCallback)
		deferredu = self.sendCommand("notcommands")
		deferredu.addCallback(self._unavailableCommandCallback)
		self._hasIdle()

	def _availableCommandCallback(self, result):
		for command in result:
			if command.has_key("command"):
				for k in self._commands.keys():
					if k[0] == command["command"]:
						format, howmany, type, keywords, available = self._commands[k]
						available = True
						self._commands[k] = format, howmany, type, keywords, available

	def _unavailableCommandCallback(self, result):
		for command in result:
			if command.has_key("command"):
				for k in self._commands.keys():
					if k[0] == command["command"]:
						format, howmany, type, keywords, available = self._commands[k]
						available = False
						self._commands[k] = format, howmany, type, keywords, available

	def _hasIdle(self):
		version = self.mpd_version.split(".")
		try:
			# check if the "idle" command is available
			# it was added on ver 0.14 (2008/12/25)
			hasidle = ((int(version[0])>0) or (int(version[1])>=14))
		except:
			print "Couldn't parse MPD-Version"
			hasidle = False
		for i in range(0, 7):
			k = ("idle", i)
			format, howmany, type, keywords, available = self._commands[k]
			self._commands[k] = format, howmany, type, keywords, hasidle

	def hasIdle(self):
		return self._commands[("idle", 0)][4]

	def _EventsToWait(self):
		return self.CallbackChanged.keys()

	def waitEvent(self, once=True):
		self._idleonce = once
		waitEventIdleDeferred = self.sendCommand("idle", self._EventsToWait())
		waitEventIdleDeferred.addCallback(self._waitEventIdleCallback)


	def _waitEventIdleCallback(self, result):
		# FIXME result can have multiple changed-entries.
		if result.has_key("changed"):
			if result["changed"] in self._EventsToWait():
				if self.CallbackChanged.has_key(result["changed"]):
					self.CallbackChanged[result["changed"]](result)
		if not self._idleonce:
			waitEventIdleDeferred = self.sendCommand("idle", self._EventsToWait())
			waitEventIdleDeferred.addCallback(self._waitEventIdleCallback)


class dictobj(dict):
	def __getattr__(self, attr):
		try:
			return self[attr]
		except KeyError:
			raise AttributeError
	def __repr__(self):
		return (object.__repr__(self).rstrip('>') + ' ..\n' +
			'  {\n    ' +
			',\n    '.join([ '%s: %s' % (k, v) for k, v in self.items() ]) +
			'\n  }>')


class MPDClientFactory(ClientFactory):
	protocol = MPDClient

	def __init__(self, password=None, ConnectCallback=None, ReadyCallback=None, DisconnectCallback=None):
		self.password=password
		self.ConnectCallback=ConnectCallback
		self.ReadyCallback=ReadyCallback
		self.DisconnectCallback=DisconnectCallback
		self._protocol = None

	def clientConnectionFailed(self, connector, reason):
		print 'connection failed:', reason.getErrorMessage()
		#self.ConnectionFailCallback()

	def clientConnectionLost(self, connector, reason):
		print 'connection lost:', reason.getErrorMessage()
		#self.ConnectionLostCallback()


	def buildProtocol(self, addr):
		self._protocol = self.protocol(self.password, self.ConnectCallback, self.ReadyCallback, self.DisconnectCallback)
		self._protocol.factory = self
		return self._protocol

	def getProtocol(self):
		return self._protocol
